import re
import emoji
from datetime import datetime
import pandas as pd
from nltk.corpus import stopwords
from wordcloud import WordCloud
import matplotlib.pyplot as plt
from collections import Counter
import seaborn as sns

# Assuming you have utility functions in a separate module
from whatsapp_analyzer import utils

class Analyzer:
    def __init__(self, chat_data):
        """
        Initializes the Analyzer with chat data (DataFrame).

        Args:
            chat_data (pd.DataFrame): DataFrame containing parsed chat data.
        """
        self.chat_data = chat_data

    def calculate_num_users(self):
        # Get unique users from the 'name' column in the DataFrame
        if "name" not in self.chat_data.columns:
            raise ValueError("Column 'name' not found in chat_data.")
        return self.chat_data["name"].nunique()


    def calculate_chat_period(self):
        """
        Calculates the start and end dates of the chat data.

        Returns:
            tuple: Start date and end date as strings.
        """
        if "date_time" not in self.chat_data.columns:
            raise ValueError("Column 'date_time' not found in chat_data.")
        
        start_date = self.chat_data["date_time"].min()
        end_date = self.chat_data["date_time"].max()
        return start_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d")


    def calculate_top_users(self, top_n=5):
        """
        Calculates the top N users by message count.

        Args:
            top_n (int): Number of top users to return.

        Returns:
            pd.DataFrame: DataFrame containing the top N users and their message counts.
        """
        if "name" not in self.chat_data.columns:
            raise ValueError("Column 'name' not found in chat_data.")
        
        if "name" in self.chat_data.columns:
            self.chat_data = self.chat_data[self.chat_data["name"].notnull()]

        user_message_counts = self.chat_data["name"].value_counts().head(top_n)
        return user_message_counts.reset_index().rename(columns={"index": "name", "name": "message_count"})

    
    def generate_wordcloud(self, column_name="message", stop_words=None):
        """Generates and displays a word cloud from the specified column."""
        text = self.chat_data[column_name].str.cat(sep=" ")
        text = re.sub(r"<Media omitted>", "", text)
        text = re.sub(r"https", "", text)  # Remove common words

        if stop_words is None:
            stop_words = set(stopwords.words("english"))
            stop_words.update(["omitted", "media", "https"])


        try:
            wordcloud = WordCloud(
                width=1600,
                height=800,
                stopwords=stop_words,
                background_color="black",
                colormap="rainbow",
            ).generate(text)

            plt.figure(figsize=(32, 18))
            plt.imshow(wordcloud, interpolation="bilinear")
            plt.axis("off")
            #plt.show()
        except ValueError as e:
            print(f"Error generating word cloud: {e}")

    def calculate_word_frequency(self, column_name="message"):
        """Calculates the frequency of words in a specified column."""
        stop_words = set(stopwords.words("english"))
        word_counts = Counter()

        for message in self.chat_data[column_name]:
            words = re.findall(r"\b\w+\b", message.lower())  # Extract words
            for word in words:
                if word not in stop_words and len(word) > 3:
                    word_counts[word] += 1

        return word_counts

    def analyze_message_length(self, column_name="mlen"):
        """Analyzes the distribution of message lengths."""
        series = self.chat_data[column_name].describe()
        return series  # Now it returns the pandas Series

    def analyze_media_count(self, column_name="mediacount"):
        """Counts the total number of media messages."""
        total_media = self.chat_data[column_name].sum()

    def analyze_media_count(self, column_name="mediacount"):
        return self.chat_data[column_name].sum() if column_name in self.chat_data else 0


    def analyze_emoji_usage(self, column_name="emojicount"):
        """Analyzes emoji usage (total count, most frequent)."""
        total_emojis = self.chat_data[column_name].sum()

        all_emojis = [e for sublist in self.chat_data["emoji"] for e in sublist]
        emoji_counts = pd.Series(all_emojis).value_counts()
        return emoji_counts
    
    def analyze_emoji_usage(self, column_name="emoji"):
        if column_name in self.chat_data:
            all_emojis = [e for sublist in self.chat_data[column_name].dropna() for e in sublist]
            emoji_counts = pd.Series(all_emojis).value_counts()
            return emoji_counts
        return pd.Series(dtype="int64")


    def create_seaborn_fig(self, x, y, sortby=None, asc=False, count=True):
        """
        Creates a Seaborn line chart for visualization.
        
        Args:
            x (str): Column name for the x-axis (e.g., 'dow' for days of the week).
            y (str): Column name for the y-axis (e.g., 'message' or 'message count').
            sortby (str): Column to sort the data by.
            asc (bool): Whether to sort in ascending order.
            count (bool): If True, count occurrences; if False, sum the values.

        Returns:
            dict: Path to the saved plot image.
        """

        try:
            # Grouping the data
            if count:
                grouped_data = self.chat_data.groupby(x, as_index=False, observed=True)[y].count()
            else:
                grouped_data = self.chat_data.groupby(x, as_index=False, observed=True)[y].sum()

            # Sorting if necessary
            if sortby:
                grouped_data = grouped_data.sort_values(by=sortby, ascending=asc)

        except Exception as e:
            print(f"An error occurred during grouping or plotting: {e}")
            traceback.print_exc()  # Print the full traceback for debugging
            return None

        # Create Seaborn line plot
        plt.figure(figsize=(10, 6))
        sns.lineplot(data=grouped_data, x=x, y=y, marker='o')
        plt.title(f"Number of {y} by {x}")
        plt.xlabel(x)
        plt.ylabel(f"Number of {y}")
        plt.xticks(rotation=45)
        plt.tight_layout()

        # Save the plot as an image
        chart_path = f'./data/{x}_vs_{y}_line_plot.png'
        plt.savefig(chart_path, bbox_inches="tight")
        plt.close()

        # Return the image path for further use in reports or as a response
        return [{
            "type": "image",
            "data": chart_path,
            "width": 800,
            "height": 400,
        }]



# whatsapp_analyzer/parser.py
import re
from dateutil import parser
import pandas as pd

class Parser:
    def __init__(self, file_path):
        """
        Initializes the Parser with the path to the WhatsApp chat file.

        Args:
            file_path (str): Path to the chat file.
        """
        self.file_path = file_path

    def parse_chat_data(self):
        """Parses the WhatsApp chat file and returns a DataFrame."""
        with open(self.file_path, "r", encoding="utf-8") as file:
            chat_lines = file.readlines()

        chat_data = []
        for line in chat_lines:
            try:
                match = re.match(
                    r"(\d{2}\/\d{2}\/\d{2}, \d{1,2}:\d{2} [apm]{2}) - (.*?): (.*)", line
                )
                if match:
                    timestamp, sender, message = match.groups()
                    date_obj = parser.parse(timestamp)
                    chat_data.append({"t": date_obj, "name": sender, "message": message})
            except (ValueError, AttributeError):
                print(f"Skipping line: {line.strip()} (Parse error)")

        return pd.DataFrame(chat_data)


# whatsapp_analyzer/stats.py

from statistics import mean, median, stdev
from collections import Counter

def calculate_message_lengths(message_lengths):
    """
    Calculates basic statistics on message lengths.
    """
    if not message_lengths:
        return {
            "mean": 0,
            "median": 0,
            "std_dev": 0,
            "max": 0,
            "min": 0
        }

    return {
        "mean": mean(message_lengths),
        "median": median(message_lengths),
        "std_dev": stdev(message_lengths),
        "max": max(message_lengths),
        "min": min(message_lengths),
    }

def calculate_active_periods(chat_data, n=5):
    """
    Finds the N most active time periods (e.g., days or hours).

    Args:
        chat_data: List of parsed chat messages.
        n: Number of top active periods to return.

    Returns:
        List of tuples: (time_period, message_count)
    """
    time_periods = Counter()
    for msg in chat_data:
        # Example: Count activity by hour
        if msg["sender"] != "System":
          time_periods[msg["date_time"].hour] += 1 

    return time_periods.most_common(n)


# whatsapp_analyzer/pdf_generator.py

import os
import pdfkit
from pathlib import Path
from bs4 import BeautifulSoup
import traceback


class PDFGenerator:
    def __init__(self, html_template_path, output_path, wkhtmltopdf_path):
        self.template = Path(html_template_path).read_text()
        self.output_path = output_path
        self.wkhtmltopdf_path = wkhtmltopdf_path
        self.classes = 'table table-sm table-bordered border-primary d-print-table fs-6'
        self.options = {
            'page-size': 'A4',
            'margin-top': '0.2in',
            'margin-right': '0.2in',
            'margin-bottom': '0.2in',
            'margin-left': '0.2in',
            'enable-local-file-access': ''
        }

    def modify_html(self, html):
        soup = BeautifulSoup(html, 'html.parser')

        if soup.body is None:
            body_tag = soup.new_tag('body')
            soup.append(body_tag)
        else:
            body_tag = soup.body

        for table in soup.find_all('table'):
            table["class"] = self.classes

        for td in soup.find_all('td'):
            td["style"] = "font-size:10px;padding:2px;text-align:center;"

        for th in soup.find_all('th'):
            th["style"] = "font-size:10px;padding:2px;text-align:left;"

        return str(soup)

    def generate_pdf(self, content, wkhtmltopdf_path=None):
        html_content = ""
        for section in content:
            if section["type"] == "html":
                html_content += section["data"]
            elif section["type"] == "table":
                html_content += section["data"].to_html(classes=self.classes, index=False)
            elif section["type"] == "image":
                width = section.get("width", 500)
                height = section.get("height", 300)
                self.options.update({'enable-local-file-access': ''})

                if not section["data"].startswith("http://") and not section["data"].startswith("https://"):
                    absolute_path = os.path.abspath(section["data"])
                    html_content += f'<img src="file:///{absolute_path}" width="{width}" height="{height}"><br>'
                else:
                    html_content += f'<img src="{section["data"]}" width="{width}" height="{height}"><br>'

        modified_html = self.modify_html(html_content)
        final_html = self.template.replace('%s', modified_html)

        html_filename = "tmp.html"
        with open(html_filename, "w", encoding="utf-8") as f:
            f.write(final_html)

        config = pdfkit.configuration(wkhtmltopdf=wkhtmltopdf_path or self.wkhtmltopdf_path)
        try:
            pdfkit.from_file(html_filename, self.output_path, options=self.options, configuration=config)
            print(f"PDF report generated: {self.output_path}")
        except Exception as e:
            print(f"Error generating PDF: {e}")
            traceback.print_exc()

        #os.remove(html_filename)  # Clean up temporary HTML file



# whatsapp_analyzer/exceptions.py

class ChatFileNotFoundError(FileNotFoundError):
    """Custom exception raised when the chat file is not found."""
    pass  # No custom behavior, just inherits FileNotFoundError

class ParseError(Exception):
    """Custom exception raised for errors during parsing."""
    pass  # No custom behavior, just inherits Exception


# tests/test_parser.py

import unittest
from whatsapp_analyzer.parser import WhatsAppParser
from whatsapp_analyzer.exceptions import ChatParseException
from datetime import datetime

class TestWhatsAppParser(unittest.TestCase):
    def test_parse_valid_line(self):
        parser = WhatsAppParser("dummy_file.txt")  # Filepath doesn't matter for this test
        line = "12/25/23, 10:00 AM - User1: Hello!"
        parser._parse_line(line)
        self.assertEqual(len(parser.data), 1)
        self.assertEqual(parser.data[0]["sender"], "User1")
        self.assertEqual(parser.data[0]["message"], "Hello!")

    def test_parse_invalid_line(self):
        parser = WhatsAppParser("dummy_file.txt")
        line = "This is not a valid chat line"
        parser._parse_line(line)
        self.assertEqual(len(parser.data), 0)  # No data should be added

    def test_parse_file_not_found(self):
        parser = WhatsAppParser("nonexistent_file.txt")
        with self.assertRaises(ChatParseException):
            parser.parse()
    
    def test_parse_valid_line_edge_cases(self):
        parser = WhatsAppParser("dummy_file.txt")
        # Test different date/time formats
        line1 = "12/25/2023, 10:00 AM - User1: Message1"
        line2 = "25/12/23, 10:00 PM - User2: Message2"
        line3 = "2023/12/25, 10:00 PM - User3: Message3"
        line4 = "12/25/23, 10:00:00 AM - User4: Message4"
        line5 = "12/25/23, 10:00 - User5: Message5"
        # ... add more variations of lines to test

        parser._parse_line(line1)
        parser._parse_line(line2)
        parser._parse_line(line3)
        parser._parse_line(line4)
        parser._parse_line(line5)
        # ... more calls to _parse_line

        self.assertEqual(len(parser.data), 5)
        self.assertEqual(parser.data[0]["date_time"], datetime(2023, 12, 25, 10, 0))
        self.assertEqual(parser.data[1]["date_time"], datetime(2023, 12, 25, 22, 0))
        self.assertEqual(parser.data[2]["date_time"], datetime(2023, 12, 25, 22, 0))
        self.assertEqual(parser.data[3]["date_time"], datetime(2023, 12, 25, 10, 0, 0))
        self.assertEqual(parser.data[4]["date_time"], datetime(2023, 12, 25, 10, 0))
        # ... add assertions for other lines

if __name__ == "__main__":
    unittest.main()


# whatsapp_analyzer/utils.py
import regex
import emoji
import pandas as pd
import numpy as np

URL_PATTERN = r"(https?://\S+)"
YOUTUBE_PATTERN = r"(https?://youtu(\.be|be\.com)\S+)"

def get_emojis(text):
    emoji_list = []
    data = regex.findall(r"\X", text)
    for word in data:
        if any(char in emoji.EMOJI_DATA for char in word):
            emoji_list.append(word)
    return emoji_list

def get_urls(text):
    url_list = regex.findall(URL_PATTERN, text)
    return url_list

def get_yturls(text):
    url_list = regex.findall(YOUTUBE_PATTERN, text)
    return url_list

def df_basic_cleanup(df):
    """
    Performs basic data cleaning and feature engineering on the DataFrame.

    Args:
        df (pd.DataFrame): DataFrame containing parsed chat data.

    Returns:
        pd.DataFrame: Cleaned DataFrame with added features.
    """
    df["date_time"] = pd.to_datetime(df["t"])
    df["date"] = df["date_time"].dt.date
    df["year"] = df["date_time"].dt.year
    df["month"] = df["date_time"].dt.month.astype(str).str.zfill(2)
    df["day"] = df["date_time"].dt.day

    df["dayn"] = df["date_time"].dt.day_name().astype("category")
    df["monthn"] = df["date_time"].dt.month_name()

    df["doy"] = df["date_time"].dt.dayofyear
    df["dow"] = df["date_time"].dt.dayofweek
    df["woy"] = df["date_time"].dt.isocalendar().week
    df["time"] = df["date_time"].dt.time
    df["hour"] = df["date_time"].dt.hour
    df["min"] = df["date_time"].dt.minute
    df["hm"] = df["hour"] + round(df["min"] / 60, 2)

    df["ym"] = df["year"].astype(str) + "-" + df["month"].astype(str)
    df["yw"] = df["year"].astype(str) + "-" + df["woy"].astype(str)
    df["yd"] = df["year"].astype(str) + "-" + df["doy"].astype(str)
    df["md"] = df["monthn"].astype(str) + "-" + df["date"].astype(str)

    df["mlen"] = df["message"].str.len()

    df["emoji"] = df["message"].apply(get_emojis)
    df["emojicount"] = df["emoji"].str.len()

    df["urls"] = df["message"].apply(get_urls)
    df["urlcount"] = df["urls"].str.len()

    df["yturls"] = df["message"].apply(get_yturls)
    df["yturlcount"] = df["yturls"].str.len()

    df["mediacount"] = np.where(df["message"] == "<Media omitted>", 1, 0)
    df["editcount"] = np.where(
        df["message"].str.contains("<This message was edited>"), 1, 0
    )
    df["deletecount"] = np.where(
        (
            (df["message"] == "This message was deleted")
            | (df["message"] == "You deleted this message")
        ),
        1,
        0,
    )

    df.drop("t", inplace=True, axis=1)
    df = df[
        [
            "date_time",
            "date",
            "year",
            "month",
            "monthn",
            "day",
            "dayn",
            "woy",
            "doy",
            "dow",
            "ym",
            "yw",
            "yd",
            "md",
            "time",
            "hour",
            "min",
            "hm",
            "name",
            "message",
            "mlen",
            "emoji",
            "emojicount",
            "urls",
            "urlcount",
            "yturls",
            "yturlcount",
            "mediacount",
            "editcount",
            "deletecount",
        ]
    ]
    return df


# whatsapp_analyzer/report_sections.py

import os
import pandas as pd
import matplotlib.pyplot as plt
from bs4 import BeautifulSoup
import emoji

def create_word_cloud_section(analyzer):
    analyzer.generate_wordcloud()
    wordcloud_filename = os.path.abspath("./data/wordcloud.png")
    plt.savefig(wordcloud_filename)
    plt.close()
    return [{"type": "image", "data": wordcloud_filename, "width": 500, "height": 300}]

def create_top_10_words_section(analyzer):
    word_frequencies = analyzer.calculate_word_frequency()
    top_10_words = word_frequencies.most_common(10)
    top_10_words_df = pd.DataFrame(top_10_words, columns=["Word", "Count"])
    return [{"type": "html", "data": "<h3>Top 10 Most Frequent Words</h3>"}, {"type": "table", "data": top_10_words_df}]

def create_message_length_section(analyzer):
    message_length_stats = analyzer.analyze_message_length()
    message_length_df = message_length_stats.reset_index()
    message_length_df.columns = ["Stat", "Value"]
    return [{"type": "html", "data": "<h3>Message Length Distribution</h3>"}, {"type": "table", "data": message_length_df}]

def create_media_count_section(analyzer):
    media_count = analyzer.analyze_media_count()
    return [{"type": "html", "data": f"<h3>Total Media Messages: {media_count}</h3>"}]

def create_top_5_emojis_section(analyzer):
    emoji_counts = analyzer.analyze_emoji_usage()
    top_5_emojis = emoji_counts.head(5)
    top_5_emojis_df = pd.DataFrame(top_5_emojis.reset_index())
    top_5_emojis_df.columns = ["Emoji", "Count"]

    # Convert the DataFrame to HTML and apply the emoji class
    html_table = top_5_emojis_df.to_html(classes="table table-sm table-bordered border-primary d-print-table fs-6", index=False)

    # Apply emoji class to any column that contains emojis
    soup = BeautifulSoup(html_table, 'html.parser')
    for td in soup.find_all('td'):
        if emoji.emoji_count(td.string):  # Check if the text contains emojis
            td['class'] = td.get('class', []) + ['emoji']

    # Convert the modified HTML back to a string
    modified_html_table = str(soup)

    return [{"type": "html", "data": "<h3>Top 5 Most Frequent Emojis</h3>"}, {"type": "html", "data": modified_html_table}]

def create_seaborn_chart_section(analyzer):
    try:
        chart_data = analyzer.create_seaborn_fig(x="dow", y="message")  # Modify this according to your data
        chart_path = chart_data[0]["data"]  # Assuming the chart image is saved and path is returned
        return [{"type": "image", "data": chart_path, "width": 800, "height": 400}]
    except ValueError as e:
        print(f"Error: {e}")
        return [{"type": "html", "data": "<p>Error: Data not found. Unable to generate chart.</p>"}]



# whatsapp_analyzer/visualizer.py

# You might want to make visualization a separate package
# that depends on your core whatsapp_analyzer package.

import matplotlib.pyplot as plt
import seaborn as sns

class ChatVisualizer:
    """
    Visualizes WhatsApp chat analysis results.
    """

    def __init__(self, analyzer):
        self.analyzer = analyzer

    def plot_message_counts(self):
        """
        Plots a bar chart of message counts per sender.
        """
        counts = self.analyzer.message_counts()
        # ... use matplotlib or seaborn to create the bar chart ...

    def plot_activity_over_time(self):
        """
        Plots a line chart of message activity over time.
        """
        activity = self.analyzer.activity_by_date()  # or activity_by_time()
        # ... use matplotlib or seaborn to create the line chart ...

    # ... other visualization methods ...